from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder,\
    XPath as ElementXPath


class SelectorBuilder(ElementSelectorBuilder):
    pass


class XPath(ElementXPath):
    # private

    @property
    def _start_string(self):
        return './' if self.adjacent is not None else './*'

    @property
    def _tag_string(self):
        if self.adjacent is not None:
            return super(XPath, self)._text_string

        return '[{} or {}]'.format(self._process_attribute('tag_name', 'th'),
                                   self._process_attribute('tag_name', 'td'))



# Translated to python from https://github.com/teamcapybara/capybara/blob/663a6003c0fa45fb83c55593183fdb2542a845ba/lib/capybara/selector/regexp_disassembler.rb

import re
from collections import OrderedDict


class RegexpDisassembler(object):

    COUNTED_REP_REGEX = r'(?P<char>.)\{(?P<min_rep>\d*)(?:,(?P<max_rep>\d*))?\}(?P<min>\?)?'
    GROUP_REGEX = r'(?P<group>\([^()]*\))' \
                  r'(?:' \
                  r'(?:' \
                  r'(?P<optional>[*?]) |' \
                  r'(?P<one_or_more>\+) |' \
                  r'(?:{})' \
                  r')\??' \
                  r')?'.format(COUNTED_REP_REGEX)

    def __init__(self, regexp):
        self.regexp = regexp
        self.pattern = self.regexp.pattern
        self._substrings = None

    @property
    def substrings(self):
        if self._substrings is not None:
            return self._substrings

        strs = []
        pattern = self.pattern
        # replace escaped characters and numbered back references with wildcard
        pattern = re.sub(r'\\.', '.', pattern)
        # replace named backreferences with wildcard
        pattern = re.sub(r'\(\?P=[^)]+\)', '.', pattern)
        # replace character classes with wildcard
        prev = ''
        while prev != pattern:
            prev = pattern
            pattern = re.sub(r'\(\?<?[=!][^)]*\)', '.', pattern)
        # remove lookahead/lookbehind assertions
        pattern = re.sub(r'\(\?<?[=!][^)]*\)', '', pattern)
        # replace named and non-matching groups with unnamed matching groups
        pattern = re.sub(r'\(\?P?(?:<[^>]+>|:)', '(', pattern)

        # simplify groups
        prev = ''
        while prev != pattern:
            prev = pattern
            pattern = re.sub(RegexpDisassembler.GROUP_REGEX,
                             lambda x: self._simplify_group(x),
                             pattern)

        # replace optional character with wildcard
        pattern = re.sub(r'.[*?]\??', '.', pattern)
        # replace one or more with character plus wildcard
        pattern = re.sub(r'(.)\+\??', r'\1.', pattern)
        # repeat counted characters
        match = re.search(RegexpDisassembler.COUNTED_REP_REGEX, pattern)
        if match is not None:
            final = match.group('char') * int(match.group('min_rep'))
            if match.group('max_rep') is not None and match.group('min') is None:
                final += '.'
            pattern = re.sub(RegexpDisassembler.COUNTED_REP_REGEX, final, pattern)

        if '|' in pattern:  # can't handle alternation here
            return []
        match = re.search(r'\A\^?(.*?)\$?\Z', pattern)
        if match is not None:
            strs = list(OrderedDict.fromkeys(x for x in match.group(1).split('.') if len(x) > 0))
            if self.regexp.flags & re.IGNORECASE:
                strs = [x.lower() for x in strs]
        return strs

    @staticmethod
    def _simplify_group(match):  # no support for alternation in groups
        if '|' in match.group('group') is not None:
            return '.'
        elif match.group('one_or_more') is not None:  # required but may repeat becomes text + wildcard
            return match.group('group')[1:-1] + '.'
        elif match.group('optional') is not None:
            return '.'
        elif match.group('min_rep') is not None:
            str = match.group('group') * int(match.group('min_rep'))
            if match.group('max_rep') is not None:
                str += '.'
            return str
        else:
            return match.group('group')[1:-1]



from nerodia.exception import LocatorException
from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder, \
    XPath as ElementXPath
from ..element.xpath_support import XpathSupport
from ...elements.text_field import TextField


class SelectorBuilder(ElementSelectorBuilder):
    pass


class XPath(ElementXPath):
    # private

    @property
    def _text_string(self):
        if self.adjacent is not None:
            return super(XPath, self)._text_string
        if 'text' in self.selector:
            self.requires_matches['text'] = self.selector.pop('text')
        return ''

    @property
    def _additional_string(self):
        if self.adjacent is not None:
            return ''
        return self._type_string(self.selector.pop('type', None))

    @property
    def _tag_string(self):
        if self.adjacent is None:
            self.selector['tag_name'] = 'input'
        return super(XPath, self)._tag_string

    def _type_string(self, typ):
        if typ is True:
            return '[{}]'.format(self._negative_type_text)
        elif typ in TextField.NON_TEXT_TYPES:
            raise LocatorException('TextField Elements can not be located by type: {}'.format(typ))
        elif typ is None:
            return '[not(@type) or ({})]'.format(self._negative_type_text)
        else:
            return '[{}]'.format(self._process_attribute('type', typ))

    @property
    def _negative_type_text(self):
        types = ['{}!={}'.format(self._lhs_for('type', lower=True), XpathSupport.escape(typ)) for
                 typ in TextField.NON_TEXT_TYPES]
        return ' and '.join(types)



from selenium.common.exceptions import InvalidElementStateException

from nerodia.exception import Error


class UserEditable(object):

    def set(self, *args):
        """
        Clear the element, then type in the given value
        :param args: value to set the input to
        """
        def func():
            self.el.clear()
            self.el.send_keys(*args)
        self._element_call(func, self.wait_for_writable)

    @property
    def value(self):
        """
        Returns value of the element
        :rtype: str
        """
        try:
            return self.attribute_value('value') or ''
        except InvalidElementStateException:
            return ''

    @value.setter  # alias
    def value(self, *args):
        self.set(*args)

    def js_set(self, *args):
        """
        Uses JavaScript to enter most of the given value
        Selenium is used to enter the first and last characters (in order to raise an error if the
        element would normally raise on sending keys)

        :param value: value to set the input to
        """
        from nerodia.exception import Error
        input_value = ''.join(args)
        if input_value:
            self.set(input_value[0])
            if hasattr(self, '_content_editable') and self._content_editable is not None:
                return self._set_content_editable(*args)
            if len(input_value) > 1:
                self._element_call(lambda: self._execute_js('setValue', self.el, input_value[:-1]))
                self.append(input_value[-1])
        value = self.value
        if value != input_value:
            raise Error("#js_set value: '{}' does not match expected input: "
                        "'{}'".format(value, input_value))

    def append(self, *args):
        """
        Appends the given value to the text in the text field
        :param args: value to add to the input
        """
        if hasattr(self, '_content_editable') and self._content_editable is not None:
            raise NotImplementedError('#append method is not supported with contenteditable '
                                      'element')
        self.send_keys(*args)

    def clear(self):
        """
        Clears the text field
        :return:
        """
        self._element_call(lambda: self.el.clear(), self.wait_for_writable)

    # private

    def _set_content_editable(self, *args):
        input_text = ''.join(args)
        self._element_call(lambda: self._execute_js('setText', self.el, input_text))
        text = self.text
        if text == input_text:
            return
        raise Error("#js_set text: '{}' does not match expected input: "
                    "'{}'".format(text, input_text))



from importlib import import_module

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder, \
    XPath as ElementXPath


class SelectorBuilder(ElementSelectorBuilder):

    def __init__(self, valid_attributes, scope_tag_name):
        self.scope_tag_name = scope_tag_name
        super(SelectorBuilder, self).__init__(valid_attributes)

    def _build_wd_selector(self, selector):
        try:
            mod = import_module(self.__module__)
            xpath = getattr(mod, 'XPath', XPath)
        except ImportError:
            xpath = XPath
        return xpath().build(selector, self.scope_tag_name)


class XPath(ElementXPath):

    def build(self, selector, scope_tag_name):
        if 'adjacent' in selector:
            return super(XPath, self).build(selector)

        index = selector.pop('index', None)

        common_string = super(XPath, self).build(selector).get('xpath')

        expressions = self._generate_expressions(scope_tag_name)
        if len(common_string) > 0:
            expressions = ["{}{}".format(e, common_string) for e in expressions]

        xpath = ' | '.join(expressions)

        if index is not None:
            xpath = self._add_index(xpath, index)

        self.selector.update(self.requires_matches)

        return {'xpath': xpath}

    # private

    @property
    def _start_string(self):
        return './' if self.adjacent is not None else ''

    @property
    def _text_string(self):
        if self.adjacent is not None:
            return super(XPath, self)._text_string

        # Can not directly locate a Row with Text because all text is in the Cells;
        # needs to use Locator#locate_matching_elements
        if 'text' in self.selector:
            self.requires_matches['text'] = self.selector.pop('text')
        return ''

    def _generate_expressions(self, scope_tag_name):
        if scope_tag_name in ['tbody', 'tfoot', 'thead']:
            return ["./*[local-name()='tr']"]
        else:
            return ["./*[local-name()='tr']",
                    "./*[local-name()='tbody']/*[local-name()='tr']",
                    "./*[local-name()='thead']/*[local-name()='tr']",
                    "./*[local-name()='tfoot']/*[local-name()='tr']"]



import re
from copy import copy
from itertools import islice
from time import sleep

import six
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.common.by import By

import nerodia
from ...exception import Error, LocatorException

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class Locator(object):
    W3C_FINDERS = {
        'css': By.CSS_SELECTOR,
        'link': By.LINK_TEXT,
        'link_text': By.LINK_TEXT,
        'partial_link_text': By.PARTIAL_LINK_TEXT,
        'tag_name': By.TAG_NAME,
        'xpath': By.XPATH
    }

    def __init__(self, query_scope, selector, selector_builder, element_validator):
        self.query_scope = query_scope  # either element or browser
        self.selector = selector
        self.selector_builder = selector_builder
        self.element_validator = element_validator
        self.normalized_selector = None
        self.driver_scope = None

    def locate(self):
        try:
            elements = self._using_selenium('first')
            return elements if elements else self._using_nerodia('first')
        except (NoSuchElementException, StaleElementReferenceException):
            return None

    def locate_all(self):
        if 'element' in self.selector:
            return [self.selector.get('element')]

        elements = self._using_selenium('all')
        return elements if elements else self._using_nerodia('all')

    # private

    def _using_selenium(self, filter='first'):
        selector = copy(self.selector)

        tag = selector.pop('tag_name', None)
        if len(self.selector) > 1:
            return

        how = list(selector)[0] if selector else 'tag_name'
        what = list(selector.values())[0] if selector else tag

        if not self._wd_is_supported(how, what, tag):
            return

        if filter == 'all':
            return self._locate_elements(how, what)
        else:
            return self._locate_element(how, what)

    def _using_nerodia(self, filter='first'):
        if 'index' in self.selector and filter == 'all':
            raise ValueError("can't locate all elements by 'index'")

        try:
            self._generate_scope()
        except LocatorException:
            return None

        built = self.selector_builder.build(self.selector)

        # Validate selector before unpacking
        self._validate_built_selector(built)
        selector, values = built

        if filter == 'all' or len(values) > 0:
            return self._locate_matching_elements(selector, values, filter)
        else:
            first_key, first_value = list(selector.items())[0]
            return self._locate_element(first_key, first_value, self.driver_scope)

    def _validate_built_selector(self, built):
        if not isinstance(built, list) or built[0] is None:
            msg = "{} was unable to build selector from {}".format(self.selector_builder.__class__,
                                                                   self.selector)
            raise LocatorException(msg)
        elif len(built) < 2 or built[1] is None:
            msg = "{}#build is not returning expected responses for the current version of " \
                  "Nerodia".format(self.selector_builder.__class__)
            raise LocatorException(msg)

    def _fetch_value(self, element, how):
        if how == 'text':
            return element.text
        elif how == 'visible':
            return element.is_displayed()
        elif how == 'visible_text':
            return element.text
        elif how == 'tag_name':
            return element.tag_name.lower()
        elif how == 'href':
            href = element.get_attribute('href')
            return href and href.strip()
        else:
            return element.get_attribute(how.replace('_', '-')) or ''

    def _matching_elements(self, elements, values, filter='first'):
        if filter == 'first':
            idx = self._element_index(elements, values)
            counter = 0

            # Generator + slice to avoid fetching values for elements that will be discarded
            matches = []
            for element in elements:
                counter += 1
                if self._matches_values(element, values) is not False:
                    matches.append(element)
            try:
                val = list(islice(matches, idx + 1))[idx]
                nerodia.logger.debug('Iterated through {} elements to locate '
                                     '{}'.format(counter, self.selector))
                return val
            except IndexError:
                return None
        else:
            nerodia.logger.debug('Iterated through {} elements to locate all '
                                 '{}'.format(len(elements), self.selector))
            return [el for el in elements if self._matches_values(el, values)]

    def _element_index(self, elements, values):
        idx = values.pop('index', 0)
        if idx < 0:
            elements.reverse()
            idx = abs(idx) - 1
        return idx

    def _generate_scope(self):
        if self.driver_scope:
            return self.driver_scope

        self.driver_scope = self.query_scope.wd

        if 'label' in self.selector:
            self._process_label('label')
        elif 'visible_label' in self.selector:
            self._process_label('visible_label')

    def _process_label(self, label_key):
        regexp = isinstance(self.selector.get(label_key), Pattern)

        if (regexp or label_key == 'visible_label') and \
                self.selector_builder.should_use_label_element:

            label = self._label_from_text(label_key)
            if not label:
                raise LocatorException("Unable to locate element with label "
                                       "{}: {}".format(label_key, self.selector.get(label_key)))

            _id = label.get_attribute('for')
            if _id:
                self.selector['id'] = _id
            else:
                self.driver_scope = label

    def _label_from_text(self, label_key):
        # TODO: this won't work correctly if @wd is a sub-element
        # TODO: figure out how to do this with find_element
        label_text = self.selector.pop(label_key, None)
        locator_key = label_key.replace('label', 'text').replace('_element', '')
        elements = self._locate_elements('tag_name', 'label', self.driver_scope)
        return next((e for e in elements if self._matches_values(e, {locator_key: label_text})),
                    None)

    def _matches_values(self, element, values):
        def check_match(how, what):
            if how == 'tag_name' and isinstance(what, six.string_types):
                return self.element_validator.validate(element, what)
            else:
                val = self._fetch_value(element, how)
                if isinstance(what, (Pattern, str)):
                    val_match = re.search(what, val) is not None
                else:
                    val_match = False
                return what == val or val_match

        matches = all(check_match(how, what) for how, what in values.items())

        if values.get('text'):
            self._text_regexp_deprecation(element, values, matches)

        return matches

    def _text_regexp_deprecation(self, element, selector, matches):
        from nerodia.elements.element import Element
        new_element = Element(self.query_scope, {'element': element})
        text_content = new_element._execute_js('getTextContent', element).strip()
        text_selector = selector.get('text', '')
        text_content_matches = re.search(text_selector, text_content) is not None
        if matches != text_content_matches:
            key = 'text' if 'text' in self.selector else 'label'
            nerodia.logger.deprecate('Using {!r} locator with RegExp: {!r} to match an element '
                                     'that includes hidden '
                                     'text'.format(key, text_selector.pattern),
                                     'visible_{}'.format(key),
                                     ids=['text_regexp'])

    def _locate_element(self, how, what, scope=None):
        scope = scope or self.query_scope.wd
        return scope.find_element(self._wd_finder(how), what)

    def _locate_elements(self, how, what, scope=None):
        scope = scope or self.query_scope.wd
        return scope.find_elements(self._wd_finder(how), what)

    def _wd_finder(self, how):
        return self.W3C_FINDERS.get(how, how)

    def _locate_matching_elements(self, selector, values, filter):
        retries = 0
        while True:
            try:
                first_key, first_value = list(selector.items())[0]
                elements = self._locate_elements(first_key, first_value, self.driver_scope) or []
                return self._matching_elements(elements, values, filter=filter)
            except StaleElementReferenceException:
                retries += 1
                sleep(0.5)
                if retries < 2:
                    continue
                target = 'element collection' if filter == 'all' else 'element'
                raise Error('Unable to locate {} from {} due to changing '
                            'page'.format(target, self.selector))

    def _wd_is_supported(self, how, what, tag):
        if how not in self.W3C_FINDERS:
            return False
        if type(what) not in nerodia._str_types:
            return False

        if how in ['partial_link_text', 'link_text', 'link']:
            nerodia.logger.deprecate('{} locator'.format(how), 'visible_text', ids=['link_text'])
            if tag in ['link', None]:
                return True
            raise Exception('Can not use {} locator to find a {} element'.format(how, what))
        elif how == 'tag_name':
            return True
        else:
            if tag:
                return False
        return True



import inspect
import re
from copy import copy
from importlib import import_module

import six
from selenium.webdriver.common.by import By

import nerodia
from nerodia.exception import LocatorException
from nerodia.locators.class_helpers import ClassHelpers
from nerodia.locators.element.regexp_disassembler import RegexpDisassembler
from nerodia.locators.element.xpath_support import XpathSupport

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern

STRING_TYPES = [six.text_type, six.binary_type]


class SelectorBuilder(object):
    VALID_WHATS = [Pattern, bool] + STRING_TYPES
    WILDCARD_ATTRIBUTE = re.compile(r'^(aria|data)_(.+)$')

    def __init__(self, valid_attributes):
        self.valid_attributes = valid_attributes
        self.custom_attributes = []
        self.xpath_builder = None
        self.selector = None

    def build(self, selector):
        rep = repr(selector)
        self.selector = selector
        self.normalize_selector()

        xpath_css = {}

        for key in copy(self.selector):
            if key in ['xpath', 'css']:
                xpath_css[key] = self.selector.pop(key)

        if len(xpath_css) == 0:
            built = self._build_wd_selector(self.selector)
        else:
            self._process_xpath_css(xpath_css)
            built = xpath_css

        if self.selector.get('index') == 0:
            self.selector.pop('index')

        nerodia.logger.debug('Converted {} to {}, with {} '
                             'to match'.format(rep, built, self.selector))
        return [built, self.selector]

    def normalize_selector(self):
        if self.selector.get('adjacent') == 'ancestor' and 'text' in self.selector:
            raise LocatorException('Can not find parent element with text locator')

        for key in self.selector.copy():
            self.check_type(key, self.selector.get(key))
            how, what = self._normalize_locator(key, self.selector.pop(key, None))
            self.selector[how] = what

    def check_type(self, how, what):
        if how in ['adjacent', 'xpath', 'css']:
            return self._raise_unless(what, str)
        elif how == 'index':
            return self._raise_unless(what, int)
        elif how == 'visible':
            return self._raise_unless(what, bool)
        elif how in ['tag_name', 'visible_text', 'text']:
            return self._raise_unless(what, 'string_or_regexp')
        elif how in ['class', 'class_name']:
            if isinstance(what, list):
                if len(what) == 0:
                    raise LocatorException("Cannot locate elements with an empty list for "
                                           "'class_name'")

                for w in what:
                    self._raise_unless(w, 'string_or_regexp')
                return
        if type(what) not in self.VALID_WHATS:
            raise TypeError(
                'expected one of {}, got {!r}:{}'.format(self.VALID_WHATS, what, what.__class__))

    @property
    def should_use_label_element(self):
        return not self._is_valid_attribute('label')

    # private

    def _normalize_locator(self, how, what):
        if how in ['tag_name', 'text', 'xpath', 'index', 'class', 'css', 'visible', 'visible_text',
                   'adjacent']:
            # include 'class' since the valid attribute is 'class_name'
            return [how, what]
        elif how == 'label':
            if self.should_use_label_element:
                return ['{}_element'.format(how), what]
            else:
                return [how, what]
        elif how == 'class_name':
            return ['class', what]
        elif how == 'caption':
            # This allows any element to be located with 'caption' instead of 'text'
            nerodia.logger.deprecate("Locating elements with 'caption'", "'text' locator",
                                     ids=['caption'])
            return ['text', what]
        else:
            self._check_custom_attribute(how)
            return [how, what]

    def _check_custom_attribute(self, attribute):
        if self._is_valid_attribute(attribute) or self.WILDCARD_ATTRIBUTE.search(attribute):
            return None
        self.custom_attributes.append(attribute)

    def _process_xpath_css(self, xpath_css):
        if len(xpath_css) > 1:
            raise LocatorException("'xpath' and 'css' cannot be combined ({})".format(xpath_css))

        if self._combine_with_xpath_or_css(self.selector):
            return

        raise LocatorException("{} cannot be combined with all of these locators "
                               "({})".format(list(xpath_css)[0], self.selector))

    # Implement this method when creating a different selector builder
    def _build_wd_selector(self, selector):
        try:
            mod = import_module(self.__module__)
            xpath = getattr(mod, 'XPath', XPath)
        except ImportError:
            xpath = XPath
        return xpath().build(selector)

    def _is_valid_attribute(self, attribute):
        return self.valid_attributes and attribute in self.valid_attributes

    @staticmethod
    def _combine_with_xpath_or_css(selector):
        keys = list(selector)
        keys = [x for x in keys if x not in ['visible', 'visible_text', 'index']]

        if len(set(keys) - {'tag_name'}) == 0:
            return True
        elif selector.get('tag_name') == 'input' and set(keys) == {'tag_name', 'type'}:
            return True
        else:
            return False

    @staticmethod
    def _raise_unless(what, typ):
        if typ == bool and isinstance(what, bool):
                return
        elif typ == 'string_or_regexp' and \
                type(what) in [six.text_type, six.binary_type, Pattern]:
            return
        elif inspect.isclass(typ):
            typ_check = (six.text_type, six.string_types) if typ == str else typ
            if isinstance(what, typ_check):
                return

        raise TypeError('expected {}, got {!r}:{}'.format(typ, what, what.__class__))


class XPath(object):
    CAN_NOT_BUILD = ['visible', 'visible_text']

    def __init__(self):
        self.selector = None
        self.requires_matches = None
        self.adjacent = None

    def build(self, selector):
        self.selector = selector

        self.requires_matches = {}

        for key in self.selector.copy():
            if key in self.CAN_NOT_BUILD:
                self.requires_matches[key] = self.selector.pop(key)

        index = self.selector.pop('index', None)
        self.adjacent = self.selector.pop('adjacent', None)

        xpath = self._start_string
        xpath += self._adjacent_string
        xpath += self._tag_string
        xpath += self._class_string
        xpath += self._text_string
        xpath += self._additional_string
        xpath += self._attribute_string

        if index is not None:
            xpath = self._add_index(xpath, index)

        self.selector.update(self.requires_matches)

        return {By.XPATH: xpath}

    # private

    def _process_attribute(self, key, value):
        if isinstance(value, Pattern):
            return self._predicate_conversion(key, value)
        else:
            return self._predicate_expression(key, value)

    def _predicate_expression(self, key, val):
        if val is True:
            return self._attribute_presence(key)
        elif val is False:
            return self._attribute_absence(key)
        else:
            return self._equal_pair(key, val)

    def _predicate_conversion(self, key, regexp):
        lower = key == 'type' or regexp.flags & re.IGNORECASE == re.IGNORECASE

        lhs = self._lhs_for(key, lower)
        results = RegexpDisassembler(regexp).substrings

        if len(results) == 0:
            self._add_to_matching(key, regexp)
            return lhs
        elif len(results) == 1 and self._starts_with(results, regexp) and not self._is_visible:
            return "starts-with({}, '{}')".format(lhs, results[0])

        self._add_to_matching(key, regexp, results)

        return ' and '.join(("contains({}, '{}')".format(lhs, substring) for substring in results))

    @property
    def _start_string(self):
        return './' if self.adjacent is not None else './/*'

    @property
    def _adjacent_string(self):
        if self.adjacent is None:
            return ''
        elif self.adjacent == 'ancestor':
            return 'ancestor::*'
        elif self.adjacent == 'preceding':
            return 'preceding-sibling::*'
        elif self.adjacent == 'following':
            return 'following-sibling::*'
        elif self.adjacent == 'child':
            return 'child::*'
        else:
            raise LocatorException('Unable to process adjacent locator with '
                                   '{}'.format(self.adjacent))

    @property
    def _tag_string(self):
        tag_name = self.selector.pop('tag_name', None)
        if tag_name is None:
            return ''
        else:
            return '[{}]'.format(self._process_attribute('tag_name', tag_name))

    @property
    def _class_string(self):
        from nerodia.locators.class_helpers import ClassHelpers
        class_name = self.selector.pop('class', None)
        if class_name is None:
            return ''

        if isinstance(class_name, str) and ' ' in class_name.strip():
            self._deprecate_class_list(class_name)

        self.requires_matches['class'] = []

        predicates = []
        for value in ClassHelpers._flatten([class_name]):
            pred = self._process_attribute('class', value)
            if pred is not None:
                predicates.append(pred)

        if len(self.requires_matches['class']) == 0:
            self.requires_matches.pop('class')

        if len(predicates) > 0:
            return '[{}]'.format(' and '.join(predicates))
        else:
            return ''

    @property
    def _text_string(self):
        text = self.selector.pop('text', None)
        if text is None:
            return ''
        elif isinstance(text, Pattern):
            self.requires_matches['text'] = text
            return ''
        else:
            return '[{}]'.format(self._predicate_expression('text', text))

    @property
    def _attribute_string(self):
        attributes = []
        for key in self.selector.copy():
            attribute = self._process_attribute(key, self.selector.pop(key))
            if attribute is not None:
                attributes.append(attribute)
        attributes = list(ClassHelpers._flatten(attributes))

        return '[{}]'.format(' and '.join(attributes)) if len(attributes) > 0 else ''

    @property
    def _additional_string(self):
        # to be used by subclasses as necessary
        return ''

    def _add_index(self, xpath, index=None):
        if self.adjacent is not None:
            return '{}[{}]'.format(xpath, index + 1)
        elif index and index >= 0 and len(self.requires_matches) == 0:
            return '({})[{}]'.format(xpath, index + 1)
        elif index and index < 0 and len(self.requires_matches) == 0:
            last_value = 'last()'
            if index < -1:
                last_value += str(index + 1)
            return '({})[{}]'.format(xpath, last_value)
        else:
            self.requires_matches['index'] = index
            return xpath

    def _deprecate_class_list(self, class_name):
        dep = "Using the 'class' locator to locate multiple classes with a String value " \
              "(i.e. '{}')".format(class_name)
        nerodia.logger.deprecate(dep, 'list (e.g. {})'.format(class_name.split()),
                                 ids=['class_list'])

    @property
    def _is_visible(self):
        return not not set(self.requires_matches).intersection(XPath.CAN_NOT_BUILD)

    def _starts_with(self, results, regexp):
        return regexp.pattern.startswith('^') and results[0] == regexp.pattern[1:]

    def _add_to_matching(self, key, regexp, results=None):
        if results is None or self._requires_matching(results, regexp):
            if key == 'class':
                self.requires_matches[key].append(regexp)
            else:
                self.requires_matches[key] = regexp

    def _requires_matching(self, results, regexp):
        if regexp.flags & re.IGNORECASE == re.IGNORECASE:
            return results[0].lower() != regexp.pattern.lower()
        else:
            return results[0] != regexp.pattern

    @staticmethod
    def _lhs_for(key, lower=False):
        if key == 'tag_name':
            return 'local-name()'
        elif key == 'href':
            return 'normalize-space(@href)'
        elif key == 'text':
            return 'normalize-space()'
        elif key == 'contains_text':
            return 'text()'
        elif isinstance(key, str):
            lhs = '@{}'.format(key.replace('_', '-'))
            return XpathSupport.lower(lhs) if lower else lhs
        else:
            raise LocatorException('Unable to build XPath using {}:{}'.format(key, key.__class__))

    def _attribute_presence(self, attribute):
        return self._lhs_for(attribute)

    def _attribute_absence(self, attribute):
        return 'not({})'.format(self._lhs_for(attribute))

    def _equal_pair(self, key, value):
        if key == 'label_element':
            # we assume 'label' means a corresponding label element, not the attribute
            text = "{}={}".format(self._lhs_for('text'), XpathSupport.escape(value))
            return "(@id=//label[{0}]/@for or parent::label[{0}])".format(text)
        elif key == 'class':
            if re.search(r'^!', value):
                value = value[1:]
                negate_xpath = True
            else:
                negate_xpath = False
            expression = "contains(concat(' ', @class, ' '), " \
                         "{})".format(XpathSupport.escape(' {} '.format(value)))
            return "not({})".format(expression) if negate_xpath else expression
        else:
            return "{}={}".format(self._lhs_for(key, key == 'type'), XpathSupport.escape(value))



from .elements.cell import Cell, CellCollection


class CellContainer(object):
    def cell(self, *args, **kwargs):
        return Cell(self, self._extract_selector(*args, **kwargs))

    def cells(self, *args, **kwargs):
        return CellCollection(self, self._extract_selector(*args, **kwargs))



from importlib import import_module


class ClassHelpers(object):
    # private

    @property
    def _locator_class(self):
        from .element.locator import Locator
        return getattr(self._import_module, 'Locator', Locator)

    @property
    def _element_validator_class(self):
        from .element.validator import Validator
        return getattr(self._import_module, 'Validator', Validator)

    @property
    def _selector_builder_class(self):
        from .element.selector_builder import SelectorBuilder
        return getattr(self._import_module, 'SelectorBuilder', SelectorBuilder)

    @property
    def _import_module(self):
        from ..module_mapping import map_module
        modules = [self.browser.locator_namespace.__name__, map_module(self._element_class_name)]
        try:
            return import_module('{}.{}'.format(*modules))
        except ImportError:
            return import_module('{}.element'.format(*modules[:1]))

    @property
    def _element_class_name(self):
        return self._element_class.__name__

    def _build_locator(self):
        from nerodia.elements.row import Row
        if self._element_class == Row:
            scope_tag_name = self.query_scope.selector.get('tag_name')
            selector_builder = self._selector_builder_class(self._element_class.ATTRIBUTES,
                                                            scope_tag_name)
        else:
            selector_builder = self._selector_builder_class(self._element_class.ATTRIBUTES)
        element_validator = self._element_validator_class()
        return self._locator_class(self.query_scope, self.selector.copy(), selector_builder,
                                   element_validator)

    @staticmethod
    def _flatten(term):
        for x in term:
            if isinstance(x, list):
                for y in x:
                    yield y
            else:
                yield x



# -*- coding: utf-8 -*-


class XpathSupport(object):
    UPPERCASE = 'ABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞŸŽŠŒ'
    LOWERCASE = 'abcdefghijklmnopqrstuvwxyzàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿžšœ'

    @staticmethod
    def escape(value):
        if "'" in value:
            parts = ["'{}'".format(part) for part in value.split("'", -1)]
            string = ",\"'\",".join(parts)

            return 'concat({})'.format(string)
        else:
            return "'{}'".format(value)

    @staticmethod
    def lower(value):
        return "translate({},'{}','{}')".format(value,
                                                XpathSupport.UPPERCASE,
                                                XpathSupport.LOWERCASE)



from nerodia.exception import LocatorException
from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder,\
    XPath as ElementXPath
from ...elements.button import Button

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class SelectorBuilder(ElementSelectorBuilder):
    pass


class XPath(ElementXPath):
    # private

    @property
    def _tag_string(self):
        if self.adjacent is not None:
            return super(XPath, self)._text_string

        # Selector builder ignores tag name and builds for both button elements and input
        # elements of type button
        self.selector.pop('tag_name', None)

        typ = self.selector.pop('type', None)
        text = self.selector.pop('text', None)

        string = '({})'.format(self._button_string(text=text, typ=typ))
        if typ is not False:
            string += ' or ({})'.format(self._input_string(text=text, typ=typ))
        return '[{}]'.format(string)

    def _button_string(self, text=None, typ=None):
        string = self._process_attribute('tag_name', 'button')
        if text is not None:
            string += ' and {}'.format(self._process_attribute('text', text))
        if typ is not None:
            string += ' and {}'.format(self._input_types(typ))
        return string

    def _input_string(self, text=None, typ=None):
        string = self._process_attribute('tag_name', 'input')
        if typ is True:
            typ = None
        string += ' and ({})'.format(self._input_types(typ))
        if text is not None:
            string += ' and {}'.format(self._process_attribute('value', text))
            self.requires_matches.pop('value', None)
        return string

    @property
    def _text_string(self):
        if self.adjacent is not None:
            return super(XPath, self)._text_string
        # text locator is already dealt with in #tag_name_string
        value = self.selector.pop('value', None)
        if value is None:
            return ''
        elif isinstance(value, Pattern):
            res = '[{} or {}]'.format(self._predicate_conversion('text', value),
                                      self._predicate_conversion('value', value))
            self.requires_matches.pop('text', None)
            return res
        else:
            return '[{} or {}]'.format(self._predicate_expression('text', value),
                                       self._predicate_expression('value', value))

    def _predicate_conversion(self, key, regexp):
        if key == 'text':
            res = super(XPath, self)._predicate_conversion('contains_text', regexp)
        else:
            res = super(XPath, self)._predicate_conversion(key, regexp)
        if 'contains_text' in self.requires_matches:
            self.requires_matches[key] = self.requires_matches.pop('contains_text')
        return res

    def _input_types(self, typ):
        if typ is None:
            types = Button.VALID_TYPES
        elif typ in Button.VALID_TYPES or typ in [True, False]:
            types = [typ]
        else:
            msg = 'Button Elements can not be located by input type: {}'.format(typ)
            raise LocatorException(msg)

        types = [self._predicate_expression('type', t) for t in types if t is not None]

        return ' or '.join(types)



import re
from copy import copy

import nerodia
from ..element.locator import Locator as ElementLocator


class Locator(ElementLocator):
    # private

    def _using_selenium(self, *args):
        return None  # force using Nerodia

    def _matches_values(self, element, values):
        if 'value' not in values:
            return super(Locator, self)._matches_values(element, values)

        cpy = copy(values)
        value = cpy.pop('value', None)

        everything_except_value = super(Locator, self)._matches_values(element, cpy)

        matches_value = re.search(value, self._fetch_value(element, 'value')) is not None
        matches_text = re.search(value, self._fetch_value(element, 'text')) is not None

        if matches_text:
            nerodia.logger.deprecate("'value' locator key for finding button text",
                                     "user 'text' locator",
                                     ids=['value_button'])

        return everything_except_value and (matches_value or matches_text)



from ..element.locator import Locator as ElementLocator


class Locator(ElementLocator):
    # private

    def _using_selenium(self, *args):
        return None  # force using Nerodia

    def _matches_values(self, element, selector):
        tag_name = None

        for key in ['text', 'value', 'label', 'visible_text']:
            if key in selector.copy():
                tag_name = tag_name or element.tag_name.lower()
                correct_key = 'value' if tag_name == 'input' else 'text'
                selector[correct_key] = selector.pop(key)

        return super(Locator, self)._matches_values(element, selector)

    def _text_regexp_deprecation(self, *args):
        return None  # does not apply to text_field
